package impl

import (
	"context"
	"fmt"
	"io/ioutil"
	"os"
	"strings"

	account "gitlab.mashroom.com/mashroom/backend/account/api"
	"gitlab.mashroom.com/mashroom/backend/common/generic/env"
	"gitlab.mashroom.com/mashroom/backend/common/generic/grpc"
	notification "gitlab.mashroom.com/mashroom/backend/notification/api"
	property "gitlab.mashroom.com/mashroom/backend/property/api"
	"golang.org/x/oauth2"
	"golang.org/x/oauth2/google"
)

const (
	IdZoopla    = "ZOOPLA"
	IdRightmove = "RIGHTMOVE"
)

var (
	EnvProjectId      = env.ResolveString("PROJECT_ID", "mashroom-email-listener")
	EnvSubjectId      = env.ResolveString("SUBJECT_ID", "portals@mashroom.com")
	EnvUrlPrefix      = env.ResolveString("URL_PREFIX", "https://testing.env.mashroom.com/property/id-")
	EnvServiceKeyFile = env.ResolveString("SERVICE_KEY_FILE", "private.json")
)

var (
	PropertyServiceUrl = env.ResolveString("PROPERTY_SERVICE_URL", "mashroom-property:10000")
	PropertyService    = property.NewPropertyServiceClient(grpc.Dial(PropertyServiceUrl, "Property"))

	NotificationServiceUrl = env.ResolveString("NOTIFICATION_SERVICE_URL", "mashroom-notification:10000")
	NotificationService    = notification.NewNotificationServiceClient(grpc.Dial(NotificationServiceUrl, "Notification"))

	AccountServiceUrl = env.ResolveString("ACCOUNT_SERVICE_URL", "mashroom-account:10000")
	AccountService    = account.NewAccountServiceClient(grpc.Dial(AccountServiceUrl, "Account"))
)

type NotificationMap struct {
	Label        string
	LabelId      string
	Topic        string
	Subscription string
}

var EnvLabels = map[string]*NotificationMap{}
var EnvHandlers = map[string]*NotificationMap{}

func init() {
	readSubscriptionsFromEnv()
}

func readSubscriptionsFromEnv() {
	for _, e := range os.Environ() {
		es := strings.SplitN(e, "=", 4)
		if len(es) != 4 {
			continue
		}

		if strings.HasPrefix(es[0], "G_") {
			fmt.Printf("%-32s:\t%v=%v=%v\n", es[0], es[1], es[2], es[3])
			name := strings.TrimPrefix(es[0], "G_")
			v := &NotificationMap{
				Label:        es[1],
				Topic:        es[2],
				Subscription: es[3],
			}
			EnvLabels[es[1]] = v
			EnvHandlers[name] = v
		}
	}
}

type StaticTokenSoruce struct {
	token *oauth2.Token
}

func serviceAccount(ctx context.Context, subject string, scopes ...string) (oauth2.TokenSource, error) {
	f, err := os.OpenFile(EnvServiceKeyFile, os.O_RDONLY, 0600)
	if err != nil {
		return nil, err
	}
	bs, err := ioutil.ReadAll(f)
	if err != nil {
		return nil, err
	}

	conf, err := google.JWTConfigFromJSON(bs, scopes...)
	if err != nil {
		return nil, err
	}
	if subject != "" {
		conf.Subject = EnvSubjectId
	}
	return conf.TokenSource(ctx), nil
}
