package impl

import (
	"context"
	"encoding/base64"
	"encoding/json"
	"io/ioutil"
	"regexp"
	"strings"

	"cloud.google.com/go/pubsub"
	"github.com/nyaruka/phonenumbers"
	admin "google.golang.org/api/admin/directory/v1"
	"google.golang.org/api/gmail/v1"
	"google.golang.org/api/option"

	account "gitlab.mashroom.com/mashroom/backend/account/api"
	"gitlab.mashroom.com/mashroom/backend/common/api/grpcerr"
	"gitlab.mashroom.com/mashroom/backend/common/generic/log"
	messaging "gitlab.mashroom.com/mashroom/backend/messaging/api"
	notification "gitlab.mashroom.com/mashroom/backend/notification/api"
	property "gitlab.mashroom.com/mashroom/backend/property/api"
)

var ErrNoProperty = grpcerr.New(grpcerr.InvalidData, "property not found")
var ErrNoUser = grpcerr.New(grpcerr.InvalidData, "user not found")
var mimeTextPlain = "text/plain"

type NotificationPayload struct {
	EmailAddress string `json:"emailAddress"`
	HistoryId    uint64 `json:"historyId"`
}

type Message struct {
	Id      string
	From    string
	Subject string
	Body    string
}

type EmailFields struct {
	Name       string
	Email      string
	Phone      string
	Address    string
	RealEstate string
	Reference  string
}

func findMessagePartBody(parts []*gmail.MessagePart, bodyData *map[string]*string) {
	if parts == nil {
		return
	}
	for _, part := range parts {
		if part.Body != nil && part.Body.Data != "" && part.MimeType != "" {
			(*bodyData)[part.MimeType] = &part.Body.Data
		}
		findMessagePartBody(part.Parts, bodyData)
	}
}

func findMessageBody(msg *gmail.Message) (*string, string) {
	bodyData := make(map[string]*string)
	if msg.Payload != nil {
		if msg.Payload.Body != nil && msg.Payload.Body.Data != "" && msg.Payload.MimeType != "" {
			bodyData[msg.Payload.MimeType] = &msg.Payload.Body.Data
		}
		if msg.Payload.Parts != nil && len(msg.Payload.Parts) > 0 {
			findMessagePartBody(msg.Payload.Parts, &bodyData)
		}
	}
	if data, ok := bodyData[mimeTextPlain]; ok {
		return data, mimeTextPlain
	}
	return nil, ""
}

func loadMessage(gm *gmail.Service, id string) (*Message, error) {
	msg, err := gm.Users.Messages.Get("me", id).Format("full").Do()
	if err != nil {
		return nil, err
	}
	var from, subject string
	for _, h := range msg.Payload.Headers {
		switch {
		case strings.EqualFold(h.Name, "from"):
			from = h.Value
		case strings.EqualFold(h.Name, "subject"):
			subject = h.Value
		}
	}
	// TODO: parse HTML body if plain text was not found?
	bodyData, _ := findMessageBody(msg)
	dat := ""
	if bodyData != nil {
		dat = strings.ReplaceAll(strings.ReplaceAll(*bodyData, "-", "+"), "_", "/")
	}
	ss, _ := ioutil.ReadAll(base64.NewDecoder(base64.StdEncoding, strings.NewReader(dat)))
	return &Message{
		Id:      id,
		From:    from,
		Subject: subject,
		Body:    string(ss),
	}, nil
}

type specificHandler func(msg *Message) *EmailFields

func processUnread(ctx context.Context, gm *gmail.Service, labelid string, entry *log.Entry, process specificHandler) {
	h, err := gm.Users.Messages.List("me").LabelIds(labelid, "UNREAD").Do()
	if err != nil {
		entry.Error("history list", err)
		return
	}
	log.Infof("%d unread message(s) for '%s' label", len(h.Messages), labelid)
	for _, m := range h.Messages {
		msg, err := loadMessage(gm, m.Id)
		if err != nil {
			log.Errorf("load message [%s] %s", m.Id, err)
			continue
		}
		if msg.Body == "" {
			log.Errorf("empty message [%s] %s", m.Id, err)
			continue
		}
		fields := process(msg)
		log.Println("FROM", msg.From, "SUBJ", msg.Subject, "RES:", fields.Name, fields.Email, fields.RealEstate)
		err = sendEventLink(ctx, fields.Email, fields.RealEstate, fields.Name, fields.Phone)
		if err != nil {
			entry.Errorf("process [%s] %s", m.Id, err)
			continue
		} else {
			_, err = gm.Users.Messages.Modify("me", m.Id, &gmail.ModifyMessageRequest{
				RemoveLabelIds: []string{"UNREAD"},
			}).Do()
			if err != nil {
				entry.Errorf("modify [%s] %s", m.Id, err)
				continue
			}
		}
	}
}

func genericHandler(process specificHandler, id string) func(gm *gmail.Service, labelid string) func(ctx context.Context, message *pubsub.Message) {
	return func(gm *gmail.Service, labelid string) func(ctx context.Context, message *pubsub.Message) {
		entry := log.WithField("source", id)
		prof, err := gm.Users.GetProfile("me").Do()
		if err != nil {
			entry.Panic(err)
		}
		last := prof.HistoryId

		processUnread(context.Background(), gm, labelid, entry, process)

		return func(ctx context.Context, message *pubsub.Message) {
			entry.Println(string(message.Data))
			payload := &NotificationPayload{}
			err := json.Unmarshal(message.Data, payload)
			if err != nil {
				entry.Error(err)
				return
			}
			processUnread(ctx, gm, labelid, entry, process)

			if payload.HistoryId > last {
				last = payload.HistoryId
			}

			message.Ack()
		}
	}
}

func sendEventLink(ctx context.Context, email, propid, fullname, phoneNumber string) error {
	prop, err := PropertyService.Get(ctx, &property.GetRequest{
		Id: propid,
	})
	if err != nil || prop.Property.GetId().GetValue() == "" {
		log.WithError(err).Error("couldn't fetch property")
		return err
	}

	user, _ := getUserByEmail(ctx, prop.Property, email)

	// convert GB phone number to E164 format
	if newPhone := FormatPhoneNumber(phoneNumber); newPhone != "" {
		phoneNumber = newPhone
	}

	_, err = NotificationService.SendNotification(ctx, &notification.SendNotificationRequest{
		Entity: notification.Entity_ENTITY_PROPERTY_QUESTION,
		Action: notification.Action_ACTION_CREATE,
		Targets: []*notification.Target{
			{
				User: user,
				Role: notification.Role_ROLE_TENANT,
				Contacts: &notification.Contacts{
					Email:    email,
					Phone:    phoneNumber,
					Fullname: fullname,
				},
			},
		},
		Embeds: []*notification.EventEmbed{
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_PROPERTY,
				EmbedValue: prop.Property.GetId().GetValue(),
			},
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_USER,
				EmbedValue: user,
			},
		},
	})
	if err != nil {
		log.WithError(err).Error("sending sms property link")
		return err
	}
	return nil
}

func getUserByEmail(ctx context.Context, p *property.Property, email string) (string, error) {
	if p == nil {
		return "", ErrNoProperty
	}

	if p.GetPeople() == nil {
		return "", ErrNoUser
	}

	people := p.GetPeople()
	if people.GetLandlord() != nil {
		acc, err := AccountService.Get(ctx, &account.GetRequest{
			Id:    people.GetLandlord().GetValue(),
			Admin: false,
		})
		if err == nil {
			if acc.Account.GetContact() != nil {
				if acc.Account.GetContactValue().GetEmailPrimaryValue() == email {
					return people.GetLandlord().GetValue(), nil
				}
			}
		}
	}

	if people.GetTenants() == nil {
		return "", ErrNoUser
	}

	for _, t := range people.GetTenants().GetValue() {
		if t.GetUser() != nil {
			acc, err := AccountService.Get(ctx, &account.GetRequest{
				Id:    t.GetUser().GetValue(),
				Admin: false,
			})
			if err == nil {
				if acc.Account.GetContact() != nil {
					if acc.Account.GetContactValue().GetEmailPrimaryValue() == email {
						return t.GetUser().GetValue(), nil
					}
				}
			}
		}
	}

	return "", ErrNoUser
}

func getOneOfRegexps(content string, rs []*regexp.Regexp) string {
	for _, r := range rs {
		if v := r.FindStringSubmatch(content); len(v) > 1 {
			res := strings.TrimSpace(v[1])
			if res != "" {
				return res
			}
		}
	}
	return ""
}

func parseEmail(rawContent string, name []*regexp.Regexp, email []*regexp.Regexp, phone []*regexp.Regexp, address []*regexp.Regexp, realestate []*regexp.Regexp, reference []*regexp.Regexp) *EmailFields {
	content := regexp.MustCompile(`=[\r\n]`).ReplaceAllString(rawContent, "")
	res := EmailFields{}
	res.Name = getOneOfRegexps(content, name)
	res.Email = getOneOfRegexps(content, email)
	res.Phone = getOneOfRegexps(content, phone)
	res.Address = getOneOfRegexps(content, address)
	res.RealEstate = getOneOfRegexps(content, realestate)
	res.Reference = getOneOfRegexps(content, reference)
	return &res
}

func zooplaHandler(msg *Message) *EmailFields {

	return parseEmail(
		msg.Body,
		[]*regexp.Regexp{regexp.MustCompile(`Name:\s*(.*?)[\r\n]`)},
		[]*regexp.Regexp{
			regexp.MustCompile(`Email:\s*(.*?)[\r\n]`),
			regexp.MustCompile(`Email address:\s*(.*?)[\r\n]`),
		},
		[]*regexp.Regexp{
			regexp.MustCompile(`Telephone:\s*(.*?)[\r\n]`),
			regexp.MustCompile(`Telephone number:\s*(.*?)[\r\n]`),
		},
		[]*regexp.Regexp{
			regexp.MustCompile(`Address:\s*(.*?)[\r\n]`),
			regexp.MustCompile(`Full address:\s*(.*?)[\r\n]`),
		},
		[]*regexp.Regexp{
			regexp.MustCompile(`Your property ref:\s*([a-z0-9-]+)[\r\n]`),
			regexp.MustCompile(`Your property ref:\s*\d+_([a-z0-9-]+)[\r\n]`),
		},
		[]*regexp.Regexp{regexp.MustCompile(`(https?://www.zoopla\.co\.uk/.*/details/\d+)`)},
	)
}

// func zooplaHandler(msg *Message) *EmailFields {
//	return parseEmail(msg.Body,
//		regexp.MustCompile(`Name:\s*(.*?)[\r\n]`),
//		regexp.MustCompile(`Email address:\s*(.*?)[\r\n]`),
//		regexp.MustCompile(`Telephone number:\s*(.*?)[\r\n]`),
//		regexp.MustCompile(`Full address:\s*(.*?)[\r\n]`),
//		regexp.MustCompile(`Your property ref:\s*\d+_([a-z0-9-]+)[\r\n]`),
//		regexp.MustCompile(`(http://www.zoopla\.co\.uk/[^/]+/details/\d+)`),
//	)
// }

func rightmoveHandler(msg *Message) *EmailFields {
	return parseEmail(msg.Body,
		[]*regexp.Regexp{regexp.MustCompile(`Name:\s*(.*?)[\r\n]`)},
		[]*regexp.Regexp{regexp.MustCompile(`Email:\s*(.*?)[\r\n]`)},
		[]*regexp.Regexp{regexp.MustCompile(`Telephone:\s*(.*?)[\r\n]`)},
		[]*regexp.Regexp{regexp.MustCompile(`(?s)Property details.*?Address:\s(.*?)[\r\n]`)},
		[]*regexp.Regexp{regexp.MustCompile(`Reference:\s\d+_([a-z0-9-]+)\s*`)},
		[]*regexp.Regexp{regexp.MustCompile(`(-)`)},
	)
}

var HandlerMap = map[string]func(gm *gmail.Service, labelid string) func(ctx context.Context, message *pubsub.Message){
	IdZoopla:    genericHandler(zooplaHandler, "zoopla"),
	IdRightmove: genericHandler(rightmoveHandler, "rightmove"),
}

func GmailSync(ctx context.Context) {
	gmailSrc, err := serviceAccount(ctx, EnvSubjectId, admin.AdminDirectoryUserReadonlyScope, admin.AdminDirectoryGroupReadonlyScope, gmail.GmailReadonlyScope, gmail.GmailModifyScope)
	if err != nil {
		log.Fatal(err)
	}
	pubSubSrc, err := serviceAccount(ctx, "", pubsub.ScopePubSub)
	if err != nil {
		log.Fatal(err)
	}
	gm, err := gmail.NewService(ctx, option.WithTokenSource(gmailSrc))
	if err != nil {
		log.Fatal(err)
	}

	ls, err := gm.Users.Labels.List("me").Do()
	if err != nil {
		log.Fatal(err)
	}
	for _, l := range ls.Labels {
		if t, ok := EnvLabels[l.Name]; ok {
			t.LabelId = l.Id
		}
	}

	pSub, err := pubsub.NewClient(ctx, EnvProjectId, option.WithTokenSource(pubSubSrc))
	if err != nil {
		log.Fatal(err)
	}
	for k, v := range EnvHandlers {
		if h, ok := HandlerMap[k]; ok {
			go func(nv *NotificationMap) {
				err = pSub.Subscription(nv.Subscription).Receive(ctx, h(gm, nv.LabelId))
				if err != nil {
					log.Fatal(err)
				}
			}(v)
		}
	}
}

func FormatPhoneNumber(phoneNumber string) string {
	num, err := phonenumbers.Parse(phoneNumber, "GB")
	if err != nil {
		return ""
	}
	return phonenumbers.Format(num, phonenumbers.E164)
}
